@testset "scaling" begin

    @testset "IdentityScaling" begin
        scaling = IdentityScaling()
        @test scaling isa Transform

        @testset "Vector" begin
            x = [1., 2., 3.]
            expected = [1., 2., 3.]

            @test FeatureTransforms.apply(x, scaling) == expected
            @test scaling(x) == expected

            @testset "Mutating" begin
                _x = copy(x)
                FeatureTransforms.apply!(_x, scaling)
                @test _x == expected
            end

            @testset "dims" begin
                @test FeatureTransforms.apply(x, scaling; dims=1) == expected
                @test_throws BoundsError FeatureTransforms.apply(expected, scaling; dims=2)
            end

            @testset "inds" begin
                @test FeatureTransforms.apply(x, scaling; inds=[2, 3]) == [2., 3.]
                @test FeatureTransforms.apply(x, scaling; dims=:, inds=[2, 3]) == [2., 3.]
                @test FeatureTransforms.apply(x, scaling; dims=1, inds=[2, 3]) == [2., 3.]
            end

            @testset "Inverse" begin
                @test FeatureTransforms.apply(x, scaling; inverse=true) == expected
            end
        end

        @testset "Matrix" begin
            M = [0.0 -0.5 0.5; 0.0 1.0 2.0]
            M_expected = [0.0 -0.5 0.5; 0.0 1.0 2.0]

           @test FeatureTransforms.apply(M, scaling) == M_expected

            @testset "Mutating" begin
                _M = copy(M)
                FeatureTransforms.apply!(_M, scaling)
                @test _M == M_expected
            end

            @testset "dims = $d" for d in (Colon(), 1, 2)
                @test FeatureTransforms.apply(M, scaling; dims=d) == M_expected
            end

            @testset "inds" begin
                @test FeatureTransforms.apply(M, scaling; inds=[2, 3]) == [0.0, -0.5]
                @test FeatureTransforms.apply(M, scaling; dims=:, inds=[2, 3]) == [0.0, -0.5]
                @test FeatureTransforms.apply(M, scaling; dims=1, inds=[2]) == [0.0 1.0 2.0]
                @test FeatureTransforms.apply(M, scaling; dims=2, inds=[2]) == reshape([-0.5; 1.0], 2, 1)
            end

            @testset "Inverse" begin
                @test FeatureTransforms.apply(M, scaling; inverse=true) == M_expected
            end
        end

        @testset "AxisArray" begin
            M = [0.0 -0.5 0.5; 0.0 1.0 2.0]
            A = AxisArray(M; foo=["a", "b"], bar=["x", "y", "z"])
            M_expected = [0.0 -0.5 0.5; 0.0 1.0 2.0]
            A_expected = AxisArray(M_expected; foo=["a", "b"], bar=["x", "y", "z"])

            @test FeatureTransforms.apply(A, scaling) == A_expected

            @testset "Mutating" begin
                _A = copy(A)
                FeatureTransforms.apply!(_A, scaling)
                @test _A isa AxisArray
                @test _A == A_expected
            end

            @testset "dims = $d" for d in (Colon(), 1, 2)
                @test FeatureTransforms.apply(M, scaling; dims=d) == A_expected
            end

            @testset "inds" begin
                @test FeatureTransforms.apply(A, scaling; inds=[2, 3]) == [0.0, -0.5]
                @test FeatureTransforms.apply(A, scaling; dims=:, inds=[2, 3]) == [0.0, -0.5]
                @test FeatureTransforms.apply(A, scaling; dims=1, inds=[2]) == [0.0 1.0 2.0]
                @test FeatureTransforms.apply(A, scaling; dims=2, inds=[2]) == reshape([-0.5; 1.0], 2, 1)
            end

            @testset "Inverse" begin
                @test FeatureTransforms.apply(A, scaling; inverse=true) == A_expected
            end
        end

        @testset "AxisKey" begin
            M = [0.0 -0.5 0.5; 0.0 1.0 2.0]
            A = KeyedArray(M; foo=["a", "b"], bar=["x", "y", "z"])
            M_expected = [0.0 -0.5 0.5; 0.0 1.0 2.0]
            A_expected = KeyedArray(M_expected; foo=["a", "b"], bar=["x", "y", "z"])

            @test FeatureTransforms.apply(A, scaling) == A_expected

            @testset "Mutating" begin
                _A = copy(A)
                FeatureTransforms.apply!(_A, scaling)
                @test _A isa KeyedArray
                @test _A == A_expected
            end

            @testset "dims = $d" for d in (Colon(), 1, 2)
                @test FeatureTransforms.apply(M, scaling; dims=d) == A_expected
            end

            @testset "inds" begin
                @test FeatureTransforms.apply(A, scaling; inds=[2, 3]) == [0.0, -0.5]
                @test FeatureTransforms.apply(A, scaling; dims=:, inds=[2, 3]) == [0.0, -0.5]
                @test FeatureTransforms.apply(A, scaling; dims=1, inds=[2]) == [0.0 1.0 2.0]
                @test FeatureTransforms.apply(A, scaling; dims=2, inds=[2]) == reshape([-0.5; 1.0], 2, 1)
            end

            @testset "Inverse" begin
                @test FeatureTransforms.apply(A, scaling; inverse=true) == A_expected
            end
        end

        @testset "NamedTuple" begin
            nt = (a = [0.0, -0.5, 0.5], b = [1.0, 0.0, 2.0])
            nt_expected = (a = [0.0, -0.5, 0.5], b = [1.0, 0.0, 2.0])

           @test FeatureTransforms.apply(nt, scaling) == [nt_expected.a, nt_expected.b]

            @testset "Mutating" begin
                _nt = deepcopy(nt)
                FeatureTransforms.apply!(_nt, scaling)
                @test _nt isa NamedTuple{(:a, :b)}
                @test _nt == nt_expected
            end

            @testset "cols = $c" for c in (:a, :b)
                @test FeatureTransforms.apply(nt, scaling; cols=[c]) ≈ [nt_expected[c]]
                @test FeatureTransforms.apply(nt, scaling; cols=c) ≈ nt_expected[c]
            end

            @testset "Inverse" begin
                @test FeatureTransforms.apply(nt, scaling; inverse=true) == [nt_expected.a, nt_expected.b]
            end
        end

        @testset "DataFrame" begin
            df = DataFrame(:a => [0.0, -0.5, 0.5], :b => [1.0, 0.0, 2.0])
            df_expected = DataFrame(:a => [0.0, -0.5, 0.5], :b => [1.0, 0.0, 2.0])

            @test FeatureTransforms.apply(df, scaling) == [df_expected.a, df_expected.b]

            @testset "Mutating" begin
                _df = deepcopy(df)
                FeatureTransforms.apply!(_df, scaling)
                @test _df isa DataFrame
                @test _df == df_expected
            end

            @testset "cols = $c" for c in (:a, :b)
                @test FeatureTransforms.apply(df, scaling; cols=[c]) ≈ [df_expected[!, c]]
                @test FeatureTransforms.apply(df, scaling; cols=c) ≈ df_expected[!, c]
            end

            @testset "Inverse" begin
                @test FeatureTransforms.apply(df, scaling; inverse=true) == [df_expected.a, df_expected.b]
            end
        end
    end

    @testset "MeanStdScaling" begin
        @testset "Constructor" begin
            x = [1., 2., 3.]

            @testset "from data" begin
                scaling = MeanStdScaling(x)

                @test scaling isa Transform
                @test scaling.mean == (all=2., )
                @test scaling.std == (all=1., )
                @test x == [1., 2., 3.]  # constructor does not mutate data
            end

            @testset "from precomputed" begin
                scaling = MeanStdScaling((all=1., ), (all=2., ))  # unlike from data
                expected = [0., 0.5, 1.]

                @test scaling isa Transform
                @test scaling.mean == (all=1., )
                @test scaling.std == (all=2., )

                @test FeatureTransforms.apply(x, scaling) == expected
            end
        end

        @testset "Vector" begin
            x = [1., 2., 3.]
            expected = [-1., 0., 1.]

            @testset "Non-mutating" begin
                scaling = MeanStdScaling(x)
                @test FeatureTransforms.apply(x, scaling) ≈ expected atol=1e-5
                @test scaling(x) ≈ expected atol=1e-5

                # Test the transform was not mutating
                @test !isapprox(x, expected; atol=1e-5)
            end

            @testset "Mutating" begin
                scaling = MeanStdScaling(x)
                _x = copy(x)
                FeatureTransforms.apply!(_x, scaling)
                @test _x ≈ expected atol=1e-5
            end

            @testset "dims" begin
                scaling = MeanStdScaling(x; dims=1)
                @test FeatureTransforms.apply(x, scaling; dims=1) == expected
                @test_throws BoundsError FeatureTransforms.apply(x, scaling; dims=2)
            end

            @testset "inds" begin
                scaling = MeanStdScaling(x)
                @test FeatureTransforms.apply(x, scaling; inds=[2, 3]) == [0., 1.]
                @test FeatureTransforms.apply(x, scaling; dims=:, inds=[2, 3]) == [0., 1.]

                scaling = MeanStdScaling(x; dims=1)
                @test FeatureTransforms.apply(x, scaling; dims=1, inds=[2, 3]) == [0., 1.]
            end

            @testset "Re-apply" begin
                scaling = MeanStdScaling(x)
                FeatureTransforms.apply(x, scaling)

                # Expect scaling parameters to be fixed to the first data applied to
                @test FeatureTransforms.apply([-0.5, 0.5, 0.0], scaling) ≈ [-2.5, -1.5, -2.0] atol=1e-5
            end

            @testset "Inverse" begin
                scaling = MeanStdScaling(x)
                transformed = FeatureTransforms.apply(x, scaling)

                @test transformed ≈ expected atol=1e-5
                @test FeatureTransforms.apply(transformed, scaling; inverse=true) ≈ x atol=1e-5
            end
        end

        @testset "Matrix" begin
            M = [0.0 -0.5 0.5; 0.0 1.0 2.0]
            M_expected = [-0.559017 -1.11803 0.0; -0.559017 0.559017 1.67705]

            @testset "Non-mutating" begin
                scaling = MeanStdScaling(M)
                @test FeatureTransforms.apply(M, scaling) ≈ M_expected atol=1e-5
                @test scaling(M) ≈ M_expected atol=1e-5

                # Test the transform was not mutating
                @test !isapprox(M, M_expected; atol=1e-5)
            end

            @testset "Mutating" begin
                scaling = MeanStdScaling(M)
                _M = copy(M)
                FeatureTransforms.apply!(_M, scaling)
                @test _M ≈ M_expected atol=1e-5
            end

            @testset "dims = :" begin
                scaling = MeanStdScaling(M; dims=:)
                @test FeatureTransforms.apply(M, scaling; dims=:) ≈ M_expected atol=1e-5
            end

            @testset "dims = 1" begin
                scaling = MeanStdScaling(M; dims=1)
                @test FeatureTransforms.apply(M, scaling; dims=1) ≈ [0.0 -0.707107 -0.707107; 0.0 0.707107 0.707107] atol=1e-5
            end

            @testset "dims = 2" begin
                scaling = MeanStdScaling(M; dims=2)
                @test FeatureTransforms.apply(M, scaling; dims=2) ≈ [0.0 -1.0 1.0; -1.0 0.0 1.0] atol=1e-5
            end

            @testset "inds" begin
                scaling = MeanStdScaling(M)
                @test FeatureTransforms.apply(M, scaling; inds=[2, 3]) ≈ [-0.559017, -1.11803] atol=1e-5
                @test FeatureTransforms.apply(M, scaling; dims=:, inds=[2, 3]) ≈ [-0.559017, -1.11803] atol=1e-5

                scaling = MeanStdScaling(M; dims=1)
                @test FeatureTransforms.apply(M, scaling; dims=1, inds=[2]) ≈ [0.0 0.707107 0.707107] atol=1e-5

                scaling = MeanStdScaling(M; dims=2)
                @test FeatureTransforms.apply(M, scaling; dims=2, inds=[2]) ≈ [-1.0 0.0]' atol=1e-5
            end

            @testset "Re-apply" begin
                scaling = MeanStdScaling(M; dims=2)
                FeatureTransforms.apply(M, scaling; dims=2)

                # Expect scaling parameters to be fixed to the first data applied to
                @test FeatureTransforms.apply([1.0 -2.0 -1.0; 0.5 0.0 0.5], scaling; dims=2) ≈ [2.0 -4.0 -2.0; -0.5 -1.0 -0.5] atol=1e-5
            end

            @testset "Inverse" begin
                scaling = MeanStdScaling(M)
                transformed = FeatureTransforms.apply(M, scaling)

                @test transformed ≈ M_expected atol=1e-5
                @test FeatureTransforms.apply(transformed, scaling; inverse=true) ≈ M atol=1e-5
            end
        end

        @testset "AxisArray" begin
            M = [0.0 -0.5 0.5; 0.0 1.0 2.0]
            A = AxisArray(M; foo=["a", "b"], bar=["x", "y", "z"])
            M_expected = [-0.559017 -1.11803 0.0; -0.559017 0.559017 1.67705]
            A_expected = AxisArray(M_expected; foo=["a", "b"], bar=["x", "y", "z"])

            @testset "Non-mutating" begin
                scaling = MeanStdScaling(A)
                @test FeatureTransforms.apply(A, scaling) ≈ A_expected atol=1e-5
                @test scaling(A) ≈ A_expected atol=1e-5

                # Test the transform was not mutating
                @test !isapprox(A, A_expected; atol=1e-5)
            end

            @testset "Mutating" begin
                scaling = MeanStdScaling(A)
                _A = copy(A)
                FeatureTransforms.apply!(_A, scaling)
                @test _A isa AxisArray
                @test _A ≈ A_expected atol=1e-5
            end

            @testset "dims = :" begin
                scaling = MeanStdScaling(A)
                @test FeatureTransforms.apply(A, scaling; dims=:) ≈ A_expected atol=1e-5
            end

            @testset "dims = 1" begin
                scaling = MeanStdScaling(A; dims=1)
                @test FeatureTransforms.apply(A, scaling; dims=1) ≈ [0.0 -0.707107 -0.707107; 0.0 0.707107 0.707107] atol=1e-5
            end

            @testset "dims = 2" begin
                scaling = MeanStdScaling(A; dims=2)
                @test FeatureTransforms.apply(A, scaling; dims=2) ≈ [0.0 -1.0 1.0; -1.0 0.0 1.0] atol=1e-5
            end

            @testset "inds" begin
                scaling = MeanStdScaling(A)
                @test FeatureTransforms.apply(A, scaling; inds=[2, 3]) ≈ [-0.559017, -1.11803] atol=1e-5
                @test FeatureTransforms.apply(A, scaling; dims=:, inds=[2, 3]) ≈ [-0.559017, -1.11803] atol=1e-5

                scaling = MeanStdScaling(A; dims=1)
                @test FeatureTransforms.apply(A, scaling; dims=1, inds=[2]) ≈ [0.0 0.707107 0.707107] atol=1e-5

                scaling = MeanStdScaling(A; dims=2)
                @test FeatureTransforms.apply(A, scaling; dims=2, inds=[2]) ≈ [-1.0 0.0]' atol=1e-5
            end

            @testset "Re-apply" begin
                scaling = MeanStdScaling(A; dims=2)
                FeatureTransforms.apply(A, scaling; dims=2)

                # Expect scaling parameters to be fixed to the first data applied to
                @test FeatureTransforms.apply([1.0 -2.0 -1.0; 0.5 0.0 0.5], scaling; dims=2) ≈ [2.0 -4.0 -2.0; -0.5 -1.0 -0.5] atol=1e-5
            end

            @testset "Inverse" begin
                scaling = MeanStdScaling(A)
                transformed = FeatureTransforms.apply(A, scaling)

                @test transformed ≈ A_expected atol=1e-5
                @test FeatureTransforms.apply(transformed, scaling; inverse=true) ≈ A atol=1e-5
            end
        end

        @testset "AxisKey" begin
            M = [0.0 -0.5 0.5; 0.0 1.0 2.0]
            A = KeyedArray(M; foo=["a", "b"], bar=["x", "y", "z"])
            M_expected = [-0.559017 -1.11803 0.0; -0.559017 0.559017 1.67705]
            A_expected = KeyedArray(M_expected; foo=["a", "b"], bar=["x", "y", "z"])

            @testset "Non-mutating" begin
                scaling = MeanStdScaling(A)
                @test FeatureTransforms.apply(A, scaling) ≈ A_expected atol=1e-5
                @test scaling(A) ≈ A_expected atol=1e-5

                # Test the transform was not mutating
                @test !isapprox(A, A_expected; atol=1e-5)
            end

            @testset "Mutating" begin
                scaling = MeanStdScaling(A)
                _A = copy(A)
                FeatureTransforms.apply!(_A, scaling)
                @test _A isa KeyedArray
                @test _A ≈ A_expected atol=1e-5
            end

            @testset "dims = :" begin
                scaling = MeanStdScaling(A)
                @test FeatureTransforms.apply(A, scaling; dims=:) ≈ A_expected atol=1e-5
            end

            @testset "dims = 1" begin
                scaling = MeanStdScaling(A; dims=1)
                @test FeatureTransforms.apply(A, scaling; dims=1) ≈ [0.0 -0.707107 -0.707107; 0.0 0.707107 0.707107] atol=1e-5
            end

            @testset "dims = 2" begin
                scaling = MeanStdScaling(A; dims=2)
                @test FeatureTransforms.apply(A, scaling; dims=2) ≈ [0.0 -1.0 1.0; -1.0 0.0 1.0] atol=1e-5
            end

            @testset "inds" begin
                scaling = MeanStdScaling(A)
                @test FeatureTransforms.apply(A, scaling; inds=[2, 3]) ≈ [-0.559017, -1.11803] atol=1e-5
                @test FeatureTransforms.apply(A, scaling; dims=:, inds=[2, 3]) ≈ [-0.559017, -1.11803] atol=1e-5

                scaling = MeanStdScaling(A; dims=1)
                @test FeatureTransforms.apply(A, scaling; dims=1, inds=[2]) ≈ [0.0 0.707107 0.707107] atol=1e-5

                scaling = MeanStdScaling(A; dims=2)
                @test FeatureTransforms.apply(A, scaling; dims=2, inds=[2]) ≈ [-1.0 0.0]' atol=1e-5
            end

            @testset "Re-apply" begin
                scaling = MeanStdScaling(A; dims=2)
                FeatureTransforms.apply(A, scaling; dims=2)

                # Expect scaling parameters to be fixed to the first data applied to
                @test FeatureTransforms.apply([1.0 -2.0 -1.0; 0.5 0.0 0.5], scaling; dims=2) ≈ [2.0 -4.0 -2.0; -0.5 -1.0 -0.5] atol=1e-5
            end

            @testset "Inverse" begin
                scaling = MeanStdScaling(A)
                transformed = FeatureTransforms.apply(A, scaling)

                @test transformed ≈ A_expected atol=1e-5
                @test FeatureTransforms.apply(transformed, scaling; inverse=true) ≈ A atol=1e-5
            end
        end

        @testset "NamedTuple" begin
            nt = (a = [0.0, -0.5, 0.5], b = [1.0, 0.0, 2.0])
            nt_expected = (a = [0.0, -1.0, 1.0], b = [0.0, -1.0, 1.0])

            @testset "Non-mutating" begin
                scaling = MeanStdScaling(nt)
                @test FeatureTransforms.apply(nt, scaling) ≈ [nt_expected.a, nt_expected.b] atol=1e-5
                @test scaling(nt) ≈ [nt_expected.a, nt_expected.b] atol=1e-5
            end

            @testset "Mutating" begin
                scaling = MeanStdScaling(nt)
                _nt = deepcopy(nt)
                FeatureTransforms.apply!(_nt, scaling)
                @test _nt isa NamedTuple{(:a, :b)}
                @test collect(_nt) ≈ collect(nt_expected) atol=1e-5
            end

            @testset "cols = $c" for c in (:a, :b)
                scaling = MeanStdScaling(nt; cols=[c])
                nt_mutated = NamedTuple{(Symbol("$c"), )}((nt_expected[c], ))
                nt_expected_ = merge(nt, nt_mutated)

                @test FeatureTransforms.apply(nt, scaling; cols=[c]) ≈ [collect(nt_expected_[c])] atol=1e-14
                @test FeatureTransforms.apply(nt, scaling; cols=c) ≈ collect(nt_expected_[c]) atol=1e-14
                @test scaling(nt; cols=[c]) ≈ [collect(nt_expected_[c])] atol=1e-14

                _nt = deepcopy(nt)
                FeatureTransforms.apply!(_nt, scaling; cols=[c])
                @test _nt isa NamedTuple{(:a, :b)}  # before applying `collect`
                @test collect(_nt) ≈ collect(nt_expected_) atol=1e-14
            end

            @testset "Re-apply" begin
                scaling = MeanStdScaling(nt)
                FeatureTransforms.apply(nt, scaling)

                # Expect scaling parameters to be fixed to the first data applied to
                nt2 = (a = [-1.0, 0.5, 0.0], b = [2.0, 2.0, 1.0])
                nt_expected2 = (a = [-2.0, 1.0, 0.0], b = [1.0, 1.0, 0.0])
                @test FeatureTransforms.apply(nt2, scaling) ≈ [nt_expected2.a, nt_expected2.b] atol=1e-5
            end

            @testset "Inverse" begin
                scaling = MeanStdScaling(nt)
                transformed = FeatureTransforms.apply(nt, scaling)
                transformed = (a = transformed[1], b = transformed[2])

                @test collect(transformed) ≈ collect(nt_expected) atol=1e-5
                @test FeatureTransforms.apply(transformed, scaling; inverse=true) ≈ [nt.a, nt.b] atol=1e-5
            end
        end

        @testset "DataFrame" begin
            df = DataFrame(:a => [0.0, -0.5, 0.5], :b => [1.0, 0.0, 2.0])
            df_expected = DataFrame(:a => [0.0, -1.0, 1.0], :b => [0.0, -1.0, 1.0])

            @testset "Non-mutating" begin
                scaling = MeanStdScaling(df)
                transformed = FeatureTransforms.apply(df, scaling)
                @test transformed ≈ [df_expected.a, df_expected.b] atol=1e-5
                @test scaling(df) ≈ [df_expected.a, df_expected.b] atol=1e-5
            end

            @testset "Mutating" begin
                scaling = MeanStdScaling(df)
                _df = deepcopy(df)
                FeatureTransforms.apply!(_df, scaling)
                @test _df isa DataFrame
                @test _df ≈ df_expected atol=1e-5
            end

            @testset "cols = $c" for c in (:a, :b)
                scaling = MeanStdScaling(df; cols=[c])

                @test FeatureTransforms.apply(df, scaling; cols=[c]) ≈ [df_expected[!, c]] atol=1e-5
                @test FeatureTransforms.apply(df, scaling; cols=c) ≈ df_expected[!, c] atol=1e-5

                _df = deepcopy(df)
                _df_expected = deepcopy(df)
                _df_expected[!, c] = df_expected[!, c]
                FeatureTransforms.apply!(_df, scaling; cols=[c])
                @test _df isa DataFrame
                @test _df ≈ _df_expected atol=1e-5
            end

            @testset "Re-apply" begin
                scaling = MeanStdScaling(df)
                FeatureTransforms.apply(df, scaling)

                # Expect scaling parameters to be fixed to the first data applied to
                df2 = DataFrame(:a => [-1.0, 0.5, 0.0], :b => [2.0, 2.0, 1.0])
                df_expected2 = DataFrame(:a => [-2.0, 1.0, 0.0], :b => [1.0, 1.0, 0.0])
                @test FeatureTransforms.apply(df2, scaling) ≈ [df_expected2.a, df_expected2.b] atol=1e-5
            end

            @testset "Inverse" begin
                scaling = MeanStdScaling(df)
                transformed = FeatureTransforms.apply(df, scaling)
                transformed = DataFrame(:a => transformed[1], :b => transformed[2])

                @test transformed ≈ df_expected atol=1e-5
                @test FeatureTransforms.apply(transformed, scaling; inverse=true) ≈ [df.a, df.b] atol=1e-5
            end
        end

        @testset "Zero std" begin
            x = ones(Float64, 3)
            expected = zeros(Float64, 3)

            scaling = MeanStdScaling(x)

            @test FeatureTransforms.apply(x, scaling) == expected  # default `eps`
            @test FeatureTransforms.apply(x, scaling; eps=1) == expected
            @test all(isnan.(FeatureTransforms.apply(x, scaling; eps=0)))  # 0/0
        end
    end
end
